"""J.A.R.V.I.S. Role Giver Cog."""
import asyncio

from dis_snek import InteractionContext, Permissions, Scale, Snake
from dis_snek.models.discord.components import ActionRow, Select, SelectOption
from dis_snek.models.discord.embed import EmbedField
from dis_snek.models.discord.role import Role
from dis_snek.models.snek.application_commands import (
    OptionTypes,
    slash_command,
    slash_option,
)
from dis_snek.models.snek.command import check, cooldown
from dis_snek.models.snek.cooldowns import Buckets

from jarvis.db.models import Rolegiver
from jarvis.utils import build_embed, get
from jarvis.utils.permissions import admin_or_permissions


class RolegiverCog(Scale):
    """J.A.R.V.I.S. Role Giver Cog."""

    def __init__(self, bot: Snake):
        self.bot = bot

    @slash_command(
        name="rolegiver", sub_cmd_name="add", sub_cmd_description="Add a role to rolegiver"
    )
    @slash_option(name="role", description="Role to add", opt_type=OptionTypes.ROLE, required=True)
    @check(admin_or_permissions(Permissions.MANAGE_GUILD))
    async def _rolegiver_add(self, ctx: InteractionContext, role: Role) -> None:
        setting = Rolegiver.objects(guild=ctx.guild.id).first()
        if setting and role.id in setting.roles:
            await ctx.send("Role already in rolegiver", ephemeral=True)
            return

        if not setting:
            setting = Rolegiver(guild=ctx.guild.id, roles=[])

        if len(setting.roles) >= 20:
            await ctx.send("You can only have 20 roles in the rolegiver", ephemeral=True)
            return

        setting.roles.append(role.id)
        setting.save()

        roles = []
        for role_id in setting.roles:
            if role_id == role.id:
                continue
            e_role = await ctx.guild.get_role(role_id)
            if not e_role:
                continue
            roles.append(e_role)
        if roles:
            roles.sort(key=lambda x: -x.position)

        value = "\n".join([r.mention for r in roles]) if roles else "None"
        fields = [
            EmbedField(name="New Role", value=f"{role.mention}"),
            EmbedField(name="Existing Role(s)", value=value),
        ]

        embed = build_embed(
            title="Rolegiver Updated",
            description="New role added to rolegiver",
            fields=fields,
        )

        embed.set_thumbnail(url=ctx.guild.icon.url)
        embed.set_author(
            name=ctx.author.display_name,
            icon_url=ctx.author.display_avatar,
        )

        embed.set_footer(text=f"{ctx.author.username}#{ctx.author.discriminator} | {ctx.author.id}")

        await ctx.send(embed=embed)

    @slash_command(
        name="rolegiver", sub_cmd_name="remove", sub_cmd_description="Remove a role from rolegiver"
    )
    @check(admin_or_permissions(Permissions.MANAGE_GUILD))
    async def _rolegiver_remove(self, ctx: InteractionContext) -> None:
        setting = Rolegiver.objects(guild=ctx.guild.id).first()
        if not setting or (setting and not setting.roles):
            await ctx.send("Rolegiver has no roles", ephemeral=True)
            return

        options = []
        for role in setting.roles:
            role: Role = await ctx.guild.get_role(role)
            option = SelectOption(label=role.name, value=str(role.id))
            options.append(option)

        select = Select(
            options=options,
            custom_id="to_delete",
            placeholder="Select roles to remove",
            min_values=1,
            max_values=len(options),
        )
        components = [ActionRow(select)]

        message = await ctx.send(content="\u200b", components=components)
        try:
            context = await self.bot.wait_for_component(
                self.bot,
                check=lambda x: ctx.author.id == x.author.id,
                messages=message,
                timeout=60 * 1,
            )
            for to_delete in context.context.values:
                setting.roles.remove(int(to_delete))
            setting.save()
            for row in components:
                for component in row["components"]:
                    component["disabled"] = True

            roles = []
            for role_id in setting.roles:
                e_role = await ctx.guild.get_role(role_id)
                if not e_role:
                    continue
                roles.append(e_role)

            if roles:
                roles.sort(key=lambda x: -x.position)

            value = "\n".join([r.mention for r in roles]) if roles else "None"
            fields = [
                EmbedField(name="Removed Role", value=f"{role.mention}"),
                EmbedField(name="Remaining Role(s)", value=value),
            ]

            embed = build_embed(
                title="Rolegiver Updated",
                description="Role removed from rolegiver",
                fields=fields,
            )

            embed.set_thumbnail(url=ctx.guild.icon.url)
            embed.set_author(
                name=ctx.author.display_name,
                icon_url=ctx.author.display_avatar,
            )

            embed.set_footer(
                text=f"{ctx.author.username}#{ctx.author.discriminator} | {ctx.author.id}"
            )

            await context.context.edit_origin(
                content=f"Removed {len(context.selected_options)} role(s)",
                embed=embed,
                components=components,
            )
        except asyncio.TimeoutError:
            for row in components:
                for component in row["components"]:
                    component["disabled"] = True
            await message.edit(components=components)

    @slash_command(name="rolegiver", sub_cmd_name="list", description="List rolegiver roles")
    async def _rolegiver_list(self, ctx: InteractionContext) -> None:
        setting = Rolegiver.objects(guild=ctx.guild.id).first()
        if not setting or (setting and not setting.roles):
            await ctx.send("Rolegiver has no roles", ephemeral=True)
            return

        roles = []
        for role_id in setting.roles:
            e_role = await ctx.guild.get_role(role_id)
            if not e_role:
                continue
            roles.append(e_role)

        if roles:
            roles.sort(key=lambda x: -x.position)

        value = "\n".join([r.mention for r in roles]) if roles else "None"

        embed = build_embed(
            title="Rolegiver",
            description=f"Available roles:\n{value}",
            fields=[],
        )

        embed.set_thumbnail(url=ctx.guild.icon.url)
        embed.set_author(
            name=ctx.author.nick if ctx.author.nick else ctx.author.username,
            icon_url=ctx.author.display_avatar,
        )

        embed.set_footer(text=f"{ctx.author.username}#{ctx.author.discriminator} | {ctx.author.id}")

        await ctx.send(embed=embed)

    @slash_command(name="role", sub_cmd_name="get", sub_cmd_description="Get a role")
    @cooldown(bucket=Buckets.USER, rate=1, interval=10)
    async def _role_get(self, ctx: InteractionContext) -> None:
        setting = Rolegiver.objects(guild=ctx.guild.id).first()
        if not setting or (setting and not setting.roles):
            await ctx.send("Rolegiver has no roles", ephemeral=True)
            return

        options = []
        for role in setting.roles:
            role: Role = await ctx.guild.get_role(role)
            option = SelectOption(label=role.name, value=str(role.id))
            options.append(option)

        select = Select(
            options=options,
            custom_id="to_delete",
            placeholder="Select roles to remove",
            min_values=1,
            max_values=len(options),
        )
        components = [ActionRow(select)]

        message = await ctx.send(content="\u200b", components=components)

        try:
            context = await self.bot.wait_for_component(
                check=lambda x: ctx.author.id == x.author_id,
                messages=message,
                timeout=60 * 5,
            )

            added_roles = []
            for role in context.context.values:
                role = await ctx.guild.get_role(int(role))
                added_roles.append(role)
                await ctx.author.add_role(role, reason="Rolegiver")

            roles = ctx.author.roles
            if roles:
                roles.sort(key=lambda x: -x.position)
                _ = roles.pop(-1)

            avalue = "\n".join([r.mention for r in added_roles]) if added_roles else "None"
            value = "\n".join([r.mention for r in roles]) if roles else "None"
            fields = [
                EmbedField(name="Added Role(s)", value=avalue),
                EmbedField(name="Prior Role(s)", value=value),
            ]

            embed = build_embed(
                title="User Given Role",
                description=f"{role.mention} given to {ctx.author.mention}",
                fields=fields,
            )

            embed.set_thumbnail(url=ctx.guild.icon.url)
            embed.set_author(
                name=ctx.author.display_name,
                icon_url=ctx.author.display_avatar,
            )

            embed.set_footer(
                text=f"{ctx.author.username}#{ctx.author.discriminator} | {ctx.author.id}"
            )

            for row in components:
                for component in row["components"]:
                    component["disabled"] = True

            await context.context.edit_origin(embed=embed, content="\u200b", components=components)
        except asyncio.TimeoutError:
            for row in components:
                for component in row["components"]:
                    component["disabled"] = True
            await message.edit(components=components)

    @slash_command(name="role", sub_cmd_name="remove", sub_cmd_description="Remove a role")
    @cooldown(bucket=Buckets.USER, rate=1, interval=10)
    async def _role_remove(self, ctx: InteractionContext) -> None:
        user_roles = ctx.author.roles

        setting = Rolegiver.objects(guild=ctx.guild.id).first()
        if not setting or (setting and not setting.roles):
            await ctx.send("Rolegiver has no roles", ephemeral=True)
            return
        elif not any(x.id in setting.roles for x in user_roles):
            await ctx.send("You have no rolegiver roles", ephemeral=True)
            return

        valid = list(filter(lambda x: x.id in setting.roles, user_roles))
        options = []
        for role in valid:
            option = SelectOption(label=role.name, value=str(role.id))
            options.append(option)

        select = Select(
            options=options,
            custom_id="to_remove",
            placeholder="Select roles to remove",
            min_values=1,
            max_values=len(options),
        )
        components = [ActionRow(select)]

        message = await ctx.send(content="\u200b", components=components)

        try:
            context = await self.bot.wait_for_component(
                check=lambda x: ctx.author.id == x.author_id,
                messages=message,
                timeout=60 * 5,
            )

            removed_roles = []
            for to_remove in context.context.values:
                role = get(user_roles, id=int(to_remove))
                await ctx.author.remove_role(role, reason="Rolegiver")
                user_roles.remove(role)
                removed_roles.append(role)

            user_roles.sort(key=lambda x: -x.position)
            _ = user_roles.pop(-1)

            value = "\n".join([r.mention for r in user_roles]) if user_roles else "None"
            rvalue = "\n".join([r.mention for r in removed_roles]) if removed_roles else "None"
            fields = [
                EmbedField(name="Removed Role(s)", value=rvalue),
                EmbedField(name="Remaining Role(s)", value=value),
            ]

            embed = build_embed(
                title="User Forfeited Role",
                description=f"{role.mention} taken from {ctx.author.mention}",
                fields=fields,
            )

            embed.set_thumbnail(url=ctx.guild.icon.url)
            embed.set_author(
                name=ctx.author.display_name,
                icon_url=ctx.author.display_avatar,
            )

            embed.set_footer(
                text=f"{ctx.author.username}#{ctx.author.discriminator} | {ctx.author.id}"
            )

            for row in components:
                for component in row["components"]:
                    component["disabled"] = True

            await context.context.edit_origin(embed=embed, components=components, content="\u200b")

        except asyncio.TimeoutError:
            for row in components:
                for component in row["components"]:
                    component["disabled"] = True
            await message.edit(components=components)

    @slash_command(
        name="rolegiver", sub_cmd_name="cleanup", description="Removed deleted roles from rolegiver"
    )
    @check(admin_or_permissions(Permissions.MANAGE_GUILD))
    async def _rolegiver_cleanup(self, ctx: InteractionContext) -> None:
        setting = Rolegiver.objects(guild=ctx.guild.id).first()
        if not setting or not setting.roles:
            await ctx.send("Rolegiver has no roles", ephemeral=True)
        guild_role_ids = [r.id for r in ctx.guild.roles]
        for role_id in setting.roles:
            if role_id not in guild_role_ids:
                setting.roles.remove(role_id)
        setting.save()

        await ctx.send("Rolegiver cleanup finished")


def setup(bot: Snake) -> None:
    """Add RolegiverCog to J.A.R.V.I.S."""
    RolegiverCog(bot)
