"""J.A.R.V.I.S. Utility Functions."""
from datetime import datetime
from operator import attrgetter
from pkgutil import iter_modules
from typing import Any, Callable, Iterable, Optional, TypeVar

import git
from dis_snek.models.discord.embed import Embed

import jarvis.cogs
import jarvis.db
from jarvis.config import get_config

__all__ = ["field", "db", "cachecog", "permissions"]

T = TypeVar("T")


def build_embed(
    title: str,
    description: str,
    fields: list,
    color: str = "#FF0000",
    timestamp: datetime = None,
    **kwargs: dict,
) -> Embed:
    """Embed builder utility function."""
    if not timestamp:
        timestamp = datetime.now()
    embed = Embed(
        title=title,
        description=description,
        color=color,
        timestamp=timestamp,
        **kwargs,
    )
    for field in fields:
        embed.add_field(**field.to_dict())
    return embed


def convert_bytesize(b: int) -> str:
    """Convert bytes amount to human readable."""
    b = float(b)
    sizes = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = 0
    while b >= 1024 and size < len(sizes) - 1:
        b = b / 1024
        size += 1
    return "{:0.3f} {}".format(b, sizes[size])


def unconvert_bytesize(size: int, ending: str) -> int:
    """Convert human readable to bytes."""
    ending = ending.upper()
    sizes = ["B", "KB", "MB", "GB", "TB", "PB"]
    if ending == "B":
        return size
    # Rounding is only because bytes cannot be partial
    return round(size * (1024 ** sizes.index(ending)))


def get_extensions(path: str = jarvis.cogs.__path__) -> list:
    """Get J.A.R.V.I.S. cogs."""
    config = get_config()
    vals = config.cogs or [x.name for x in iter_modules(path)]
    return ["jarvis.cogs.{}".format(x) for x in vals]


def update() -> int:
    """J.A.R.V.I.S. update utility."""
    repo = git.Repo(".")
    dirty = repo.is_dirty()
    current_hash = repo.head.object.hexsha
    origin = repo.remotes.origin
    origin.fetch()
    if current_hash != origin.refs["main"].object.hexsha:
        if dirty:
            return 2
        origin.pull()
        return 0
    return 1


def get_repo_hash() -> str:
    """J.A.R.V.I.S. current branch hash."""
    repo = git.Repo(".")
    return repo.head.object.hexsha


def find(predicate: Callable[[T], Any], seq: Iterable[T]) -> Optional[T]:
    for element in seq:
        if predicate(element):
            return element
    return None


def get(iterable: Iterable[T], **attrs: Any) -> Optional[T]:
    if len(attrs) == 1:
        k, v = attrs.popitem()
        pred = attrgetter(k.replace("__", "."))
        for elem in iterable:
            if pred(elem) == v:
                return elem
        return None

    converted = [(attrgetter(attr.replace("__", ".")), value) for attr, value in attrs.items()]

    for elem in iterable:
        if all(pred(elem) == value for pred, value in converted):
            return elem
        return None
